# Skip Tcl failures
import matplotlib

matplotlib.use("Agg", force=True)

from tests.fixtures.fixture_regridder import (
    disk,
    disk_layered,
    expected_results_centroid,
    expected_results_linear,
    expected_results_overlap,
    grid_data_a,
    grid_data_a_1d,
    grid_data_a_2d,
    grid_data_a_layered,
    grid_data_a_layered_1d,
    grid_data_a_layered_2d,
    grid_data_b,
    grid_data_b_1d,
    grid_data_b_2d,
    grid_data_b_flipped_1d,
    grid_data_c,
    grid_data_c_1d,
    grid_data_c_2d,
    grid_data_d,
    grid_data_d_1d,
    grid_data_dask_expected,
    grid_data_dask_expected_layered,
    grid_data_dask_source,
    grid_data_dask_source_layered,
    grid_data_dask_target,
    grid_data_e,
    grid_data_e_1d,
    quads_0_25,
    quads_1,
    quads_structured,
)

__all__ = (
    "disk",
    "disk_layered",
    "expected_results_centroid",
    "expected_results_linear",
    "expected_results_overlap",
    "grid_data_a",
    "grid_data_a_1d",
    "grid_data_a_2d",
    "grid_data_a_layered",
    "grid_data_a_layered_1d",
    "grid_data_a_layered_2d",
    "grid_data_b",
    "grid_data_b_1d",
    "grid_data_b_2d",
    "grid_data_b_flipped_1d",
    "grid_data_c",
    "grid_data_c_1d",
    "grid_data_c_2d",
    "grid_data_d",
    "grid_data_d_1d",
    "grid_data_dask_expected",
    "grid_data_dask_expected_layered",
    "grid_data_dask_source",
    "grid_data_dask_source_layered",
    "grid_data_dask_target",
    "grid_data_e",
    "grid_data_e_1d",
    "quads_0_25",
    "quads_1",
    "quads_structured",
)
